"""
Constants for authorization mocking
"""

MOCK_ACCOUNT_ID = "81a6dd2d-160e-4530-864d-e7ab524462a0"

MOCK_TOKEN = ("Bearer eyJraWQiOiIzUlRMQ1l1MFJWZ09LMlZSWWZ1S1dSVnE2WWhQa2JMT3BHM2o2ejArRDhJPSIsImFsZyI6IlJTMjU2In0."
              "eyJhdF9oYXNoIjoiZm5WdnRKYkpXdFJWRnhiSWVkNVU3USIsInN1YiI6IjgxYTZkZDJkLTE2MGUtNDUzMC04NjRkLWQ3YWI1MjQ0Nj"
              "JhMCIsImNvZ25pdG86Z3JvdXBzIjpbInVzLXdlc3QtMl9STm9kVGt1TkNfRmFjZWJvb2siXSwiaXNzIjoiaHR0cHM6XC9cL2NvZ2"
              "5pdG8taWRwLnVzLXdlc3QtMi5hbWF6b25hd3MuY29tXC91cy13ZXN0LTJfUk5vZFRrdU5DIiwiY29nbml0bzp1c2VybmFtZSI6"
              "IkZhY2Vib29rXzI2NTI4MTQ1MDk4MzAzMSIsImdpdmVuX25hbWUiOiJNaXNoZWwiLCJhdWQiOiI2MzFuY2pibmVsazg1NGtv"
              "amphcGo1NTN0aCIsImlkZW50aXRpZXMiOlt7InVzZXJJZCI6IjI2NTI4MTQ1MDk4MzAzMSIsInByb3ZpZGVyTmFtZSI6IkZh"
              "Y2Vib29rIiwicHJvdmlkZXJUeXBlIjoiRmFjZWJvb2siLCJpc3N1ZXIiOm51bGwsInByaW1hcnkiOiJ0cnVlIiwiZGF0ZUN"
              "yZWF0ZWQiOiIxNTM2MzIwMDM1MTI3In1dLCJ0b2tlbl91c2UiOiJpZCIsImF1dGhfdGltZSI6MTUzODU1NDc2MywibmFtZSI6"
              "Ik1pc2hlbCBDcnVpc2UiLCJleHAiOjE1Mzg1NTgzNjMsImlhdCI6MTUzODU1NDc2MywiZmFtaWx5X25hbWUiOiJDcnVpc2UiLC"
              "JlbWFpbCI6InRlc3R1c2VyOTUxYmJAZ21haWwuY29tIn0.YMxi4xXdHgWikWHNPh-sr-gFNGTMlva4VJWCagGrIone6YTVX4uR"
              "gHK-du06nJ-bqqaZBTzZOwaV2NFp9oQDOI5TzMykYBskAEJ0CiI1YP_nUq8KcZRtElWPk6IbGtHhG2GqYxHedvCZxwKalWxi01"
              "iGnAayNbNA9-NaqPoXu_9J2VKC24lgbR7JEifwKBHzq37SmV20kyKeMVEzN7XxyQgoyUGr2ndfn3XT289_9MpaUNaaq66QxGMB"
              "UIC8jkqFGauZc7iCVxLuV79EFIX_Y4nGh95qNhBc8hQb4_TNhpSN2rnqk_ePypBcL54S26bWJgBxGr8bebDFM9q5OttjDg")

STANDARD_AUTHORIZER_CONTEXT = {
    "requestContext": {
        "authorizer": {
            "claims": {
                "at_hash": "gnVvtJbJWtRVFxbIed3U7Q",
                "aud": "731ncjbnelk854kijjapj553th",
                "auth_time": "1538554763", "cognito:groups":
                "us-west-2_RNodTkuNC_Facebook",
                "cognito:username": "Facebook_9876543210",
                "email": "testuser123456@gmail.com",
                "exp": "Wed Oct 03 09:19:23 UTC 2018",
                "family_name": "User",
                "given_name": "Test",
                "iat": "Wed Oct 03 08:19:23 UTC 2018",
                "identities": ("{\"dateCreated\":\"1536320035127\",\"userId\":\"265281450983031\","
                               " \"providerName\":""\"Facebook\",\"providerType\":\"Facebook\","
                               " \"issuer\":null,\"primary\":\"true\"}"),
                "iss": "https://cognito-idp.us-west-2.amazonaws.com/us-west-2_WAVELENGTH_ACCOUNT",
                "name": "Test User",
                "sub": MOCK_ACCOUNT_ID,
                "token_use": "id"
            }
        }
    }
}



# pylint: disable=W0212
# pylint: disable=C0103
# pylint: disable=W0603
# pylint: disable=C0411
from http.client import responses

import requests

from wavelength_test.integration.util.event_to_request import (VERB_GET,
                                                        VERB_POST, VERB_PUT,
                                                        VERB_PATCH, VERB_DELETE)
from wavelength_test.integration.util.event_to_request import (update_request_event,
                                                        check_headers, VERB_INVOKE)
from wavelength_test.integration.util.invoker_base import RESPONSE_CONVERTERS, try_json
from wavelength_test.integration.util.mock_constants import (MOCK_TOKEN,
                                                      MOCK_ACCOUNT_ID)


class InvokerResponseLocal:
    """
    InvokerResponse for integration tests against local
    """

    def __init__(self, httpResponse):
        """
        InvokerResponseLocal init, sets various attributes as follows:
        """
        self.response = httpResponse

        body = self.response['body']
        content_type = self.response['headers'].get('Content-Type', 'undefined')
        # Default to try_json otherwise we could break many tests since most responses aren't
        # returning Content-Type header from the lib's get_standard_response (but they should -BM)
        converter = RESPONSE_CONVERTERS.get(content_type, try_json)

        self.response_body = converter(body)
        self.header = self.response['headers']
        self.status_code = self.response['statusCode']

    def __getitem__(self, key):
        """
        This method ensures that the pre-existing contract stays working.
        it overwrites the bracket syntax.
        """
        return self.response_body[key]


class InvokerLocal:
    """
    Invoker for local integration testing
    """

    def __init__(self, local_invoke_mapper, ctx):
        """
        Initializes InvokerLocal
        :param local_invoke_mapper:
        """
        self._ctx = ctx
        self._local_invoke_mapper = local_invoke_mapper
        self._cognito_id_token = MOCK_TOKEN

    @staticmethod
    def get_decrypted_cognito_profile_id(profile_num=1):
        """
        Gets the cognito Account ID
        """
        return f'{MOCK_ACCOUNT_ID}-{profile_num}'

    def get_cognito_id_token(self, entitlements: [str]):
        """
        Gets the cognito JWT, ignores entitlements for now
        """
        return self._cognito_id_token

    def get(self, event, resource, **kwargs):
        """
        Get request
        """
        event = update_request_event(event, self._cognito_id_token)
        result = self._local_invoke_mapper[resource][VERB_GET](
            event, self._ctx)
        return self._process_response(result, **kwargs)

    def delete(self, event, resource, **kwargs):
        """
        Delete request
        """
        event = update_request_event(event, self._cognito_id_token)
        result = self._local_invoke_mapper[resource][VERB_DELETE](
            event, self._ctx)
        return self._process_response(result, **kwargs)

    def post(self, event, resource, **kwargs):
        """
        Post request
        """
        event = update_request_event(event, self._cognito_id_token)
        result = self._local_invoke_mapper[resource][VERB_POST](
            event, self._ctx)
        return self._process_response(result, **kwargs)

    def put(self, event, resource, **kwargs):
        """
        Put request
        """
        event = update_request_event(event, self._cognito_id_token)
        result = self._local_invoke_mapper[resource][VERB_PUT](
            event, self._ctx)
        return self._process_response(result, **kwargs)

    def patch(self, event, resource, **kwargs):
        """
        Put request
        """
        event = update_request_event(event, self._cognito_id_token)
        result = self._local_invoke_mapper[resource][VERB_PATCH](
            event, self._ctx)
        return self._process_response(result, **kwargs)

    def invoke(self, event):
        """
        Direct Lambda invocation
        """
        return self._local_invoke_mapper['default'][VERB_INVOKE](event, self._ctx)

    @classmethod
    def _process_response(cls, response, **kwargs):
        """
        Process the response (get, delete, post, put, etc)
        :param response:
        :return:
        """
        if not response:
            return response
        if response['statusCode'] in range(200, 499):
            check_headers(response['headers'])
            if response['statusCode'] in range(200, 399):
                return InvokerResponseLocal(response)
        response_model = requests.models.Response()
        response_model.code = responses.get(response['statusCode'])
        response_model.error_type = "HTTP Exception"
        response_model.status_code = response['statusCode']
        response_model._content = response['body'].encode()
        raise requests.exceptions.HTTPError(
            'LocalInvoker HTTP Error', response=response_model)



# pylint: disable=C0103
import base64
import json
import os

import boto3
import requests

from wavelength_test.integration.util.aws_utils import get_api_key, get_cognito_id_token
from wavelength_test.integration.util.event_to_request import (check_headers,
                                                        VERB_GET,
                                                        VERB_POST, VERB_PUT,
                                                        VERB_PATCH, VERB_DELETE)
from wavelength_test.integration.util.invoker_base import RESPONSE_CONVERTERS, try_json
from wavelength_test.integration.util.event_to_request import event_to_request_url, event_to_request_headers, event_to_body

ENDPOINT_TYPE_LAMBDA = 'lambda'
ENDPOINT_TYPE_GATEWAY = 'apigateway'


class InvokerResponseAWS:
    """
    InvokerResponse for integration tests against AWS
    """

    def __init__(self, httpResponse):
        """
        InovkerAWS init, sets various attributes as follows:
        """
        self.response = httpResponse

        content_type = self.response.headers.get('Content-Type', 'undefined')
        converter = RESPONSE_CONVERTERS.get(content_type, try_json)

        self.response_body = converter(self.response.text)
        self.header = self.response.headers
        self.status_code = self.response.status_code

    def __getitem__(self, key):
        """
        This method ensures that the pre-existing contract stays working.
        it overwrites the bracket syntax.
        """
        return self.response_body[key]


class InvokerAWS:
    """
    Invoker for integration tests against AWS
    """
    _cognito_id_token_val = None

    def __init__(self, service_name, stage, api_version, **kwargs):
        """
        InvokerAWS init
        :param service_name:
        :param stage:
        :param api_version:
        :param endpoint_type: [lambda,apigateway]
        :param resource: path in addition to an expected base
        """
        self._service_name = service_name
        self._stage = stage
        self._api_version = api_version
        self._region = os.environ.get('region', 'us-east-1')
        self._ssm_client = boto3.client('ssm', self._region)
        self._gateway_client = boto3.client('apigateway', self._region)
        self._api_key = get_api_key(
            service_name, self._ssm_client, self._gateway_client)
        self._endpoint_type = kwargs.get('endpoint_type')
        self._resource = kwargs.get('resource')

        if self._resource:
            param_name = (
                f'/{self._endpoint_type}/{self._stage}/{service_name}/v{self._api_version}/{self._resource}')
        else:
            param_name = (
                f'/{self._endpoint_type}/{self._stage}/{service_name}/v{self._api_version}')
        response = self._ssm_client.get_parameter(Name=param_name)
        self._request_url = response['Parameter']['Value']

    def get_decrypted_cognito_profile_id(self, profile_num=1):
        """
        Decodes the cognito JWT, gets the IDToken at [1], and adds '-1' to the end
        :return: decrypted JWT IDToken with '-{profile_num}' on the end
        """
        tokens = self._cognito_id_token.split('.')
        decoded_token = json.loads(base64.b64decode(tokens[1]).decode('utf-8'))
        sub_from_token = decoded_token['sub'] + f'-{profile_num}'
        return sub_from_token

    def get_cognito_id_token(self):
        """
        Gets the cognito JWT
        """
        return self._cognito_id_token

    @property
    def _cognito_id_token(self):
        return None
        # TODO - use JWT service
        if not type(self)._cognito_id_token_val:
            type(self)._cognito_id_token_val = get_cognito_id_token(self._service_name,
                                                                    self._region,
                                                                    self._ssm_client)
        return type(self)._cognito_id_token_val

    def _get_url(self, resource):
        """
        Get request url
        :param resource:
        :return:
        """
        paths = self._request_url.split('/')
        if resource and resource != 'default' and resource not in paths:
            url = self._request_url + '/' + resource
        else:
            url = self._request_url
        return url

    def get(self, event, resource, **kwargs):
        """
        Get
        :param event:
        :param resource:
        :return:
        """
        url = self._get_url(resource)
        if self._endpoint_type == ENDPOINT_TYPE_LAMBDA:
            raise Exception(
                'Bad invocation type change invoker type to apigateway')
        return self._invoke_api(VERB_GET, url, self._api_key, event, **kwargs)

    def delete(self, event, resource, **kwargs):
        """
        Delete
        :param event:
        :param resource:
        :return:
        """
        url = self._get_url(resource)
        if self._endpoint_type == ENDPOINT_TYPE_LAMBDA:
            raise Exception(
                'Bad invocation type change invoker type to apigateway')
        return self._invoke_api(VERB_DELETE, url, self._api_key, event, **kwargs)

    def post(self, event, resource, **kwargs):
        """
        Post
        :param event:
        :param resource:
        :return:
        """
        url = self._get_url(resource)
        if self._endpoint_type == ENDPOINT_TYPE_LAMBDA:
            raise Exception(
                'Bad invocation type change invoker type to apigateway')
        return self._invoke_api(VERB_POST, url, self._api_key, event, **kwargs)

    def put(self, event, resource, **kwargs):
        """
        Put
        :param event:
        :param resource:
        :return:
        """
        url = self._get_url(resource)
        if self._endpoint_type == ENDPOINT_TYPE_LAMBDA:
            raise Exception(
                'Bad invocation type change invoker type to apigateway')
        return self._invoke_api(VERB_PUT, url, self._api_key, event, **kwargs)

    def patch(self, event, resource, **kwargs):
        """
        Put
        :param event:
        :param resource:
        :return:
        """
        url = self._get_url(resource)
        if self._endpoint_type == ENDPOINT_TYPE_LAMBDA:
            raise Exception(
                'Bad invocation type change invoker type to apigateway')
        return self._invoke_api(VERB_PATCH, url, self._api_key, event, **kwargs)

    def invoke(self, event):
        """
        Direct Lambda invocation
        """
        if self._endpoint_type == ENDPOINT_TYPE_GATEWAY:
            raise Exception(
                'Bad invocation type change invoker type to lambda')
        return self._invoke_lambda(self._request_url, event)

    def _invoke_api(self, verb, url, apikey, event, **kwargs):
        """
        Invokes API
        :param verb:
        :param url:
        :param apikey:
        :param event:
        :param content_type:
        :return: response
        """
        content_type = kwargs.get('content_type', 'application/json')
        request_url = event_to_request_url(url, event)
        body = event_to_body(event)
        should_redirect = kwargs.get('allow_redirect', True)
        if body:
            response = getattr(requests, verb)(request_url,
                                               headers=event_to_request_headers(event,
                                                                                content_type,
                                                                                apikey,
                                                                                self._cognito_id_token),
                                               json=body, allow_redirects=should_redirect)
        else:
            response = getattr(requests, verb)(request_url,
                                               headers=event_to_request_headers(event,
                                                                                content_type,
                                                                                apikey,
                                                                                self._cognito_id_token),
                                               allow_redirects=should_redirect)

        if response.status_code in range(200, 499) and not response.status_code == 400:
            # API Gateway does not return CORS headers on validation failure
            check_headers(response.headers)
        response.raise_for_status()
        return InvokerResponseAWS(response)

    @classmethod
    def _invoke_lambda(cls, arn, event):
        """
        Direct Lambda invocation
        :param arn:
        :param event:
        :return: payload
        """
        lambda_client = boto3.client(ENDPOINT_TYPE_LAMBDA)
        res = lambda_client.invoke(
            FunctionName=arn,
            InvocationType='RequestResponse',
            Payload=json.dumps(event),
        )
        return json.loads(res['Payload'].read().decode("utf-8"))



"""
Placing common invoker functions here since lambda_util imports local and aws invokers so
importing lambda_util from either of those will create a circular ref
"""
import json
from typing import Union, Any


def try_json(source: str) -> Union[dict, str]:
    """
    Tries to coerce JSON from a body string if not None, returns original value on fail
    :param source:
    :return: Union[dict, str]
    """
    if not isinstance(source, str):
        return source
    try:
        return json.loads(source)
    except ValueError:
        return source


def pass_through(source: Any) -> Any:
    """
    Pass-through converter for unhandled mime-types
    :param source: Any
    :return: Any
    """
    return source


RESPONSE_CONVERTERS = {
    'text/html': pass_through,
    'text/plain': pass_through,
    'application/xml': pass_through
}



# pylint: disable=C0103

import os

from botocore.exceptions import ClientError
from typing import List

from wavelength_test.integration.util.event_to_request import VERBS
from wavelength_test.integration.util.lambda_invoker_aws import InvokerAWS
from wavelength_test.integration.util.lambda_invoker_local import InvokerLocal

EXPECTED_HEADERS = {
    "Access-Control-Allow-Origin": "*",
    "Access-Control-Allow-Credentials": "true"
}
SERVERLESS = None


class ServerlessYMLInjector:
    """
    Used to inject the serverless.yml file into this file for the Invoker __init__'s service_name attribute
    !Important: the serverless.yml file must be opened and injected BEFORE using the Invoker.
    This should be done in the setUp() of the subclassed integration_test_base.py in each project as follows:
    with open('./serverless.yml', "r") as f:
        serverless_yaml = yaml.load(f)
    ServerlessYMLInjector(serverless_yaml)
    """

    def __init__(self, serverless_yaml):
        global SERVERLESS  # pylint: disable=global-statement
        SERVERLESS = serverless_yaml


class Context:
    """
    Creates Mock Context object
    """

    def __init__(self):
        self.function_name = 'stage-func_name'
        self.aws_request_id = 'test'
        self.invoked_function_arn = 'test_url'


class LocalInvokeMapper:
    """
    Mapper for Local Invoker
    """

    def __init__(self, verb, func, resource=None):
        if verb.lower() not in VERBS:
            raise Exception('Invalid verb must be one of {}'.format(VERBS))
        self._resource = resource
        self._func = func
        self._verb = verb.lower()

    def get_resource(self):
        """
        Gets resource
        """
        return self._resource

    def get_verb(self):
        """
        Gets verb
        """
        return self._verb

    def get_function(self):
        """
        Gets function
        """
        return self._func


class Invoker:
    """
    Invoker determines which resources to use if local or AWS
    """

    def __init__(self, mappers: List[LocalInvokeMapper], endpoint_type='apigateway', resource=None):
        """
        Create an invoker per endpoint

        service name - this project is stdemo
        mappers - map an optional resource (multiple endpoint paths) and verbs to a local 'lambda' function
        endpoint_type - is the method to call api gateway or lambda
        resource - optional extra path parameter for when reading ssm parameters for endpoints off of the base
        """
        if SERVERLESS is not None:
            self._service_name = SERVERLESS['custom']['service_name']
        else:
            raise Exception(
                'You must inject serverless.yml file using ServerlessYMLInjector before using the Invoker')

        self._stage = os.environ.get('stageName')
        self._api_version = os.environ.get('api_version', 1)
        self._endpoint_type = endpoint_type
        self._local_invoke_mapper = dict()
        for mapper in mappers:
            self._local_invoke_mapper.setdefault(mapper.get_resource() if mapper.get_resource(
            ) else 'default', {}).update({mapper.get_verb(): mapper.get_function()})
        try:
            self.invoker = InvokerAWS(self._service_name,
                                      self._stage,
                                      self._api_version,
                                      endpoint_type=self._endpoint_type,
                                      resource=resource) \
                if self._stage else InvokerLocal(self._local_invoke_mapper, Context())
        except ClientError as error:
            boto_error = _get_boto3_error(error)
            if boto_error.get('aws', {}).get('code', '') == 'ParameterNotFound':
                raise Exception(
                    f'One or more SSM parameters are missing for stage {self._stage}') from error
            if 'The security token included in the request is expired' in boto_error.get('message', ''):
                raise Exception(f'Failed to execute AWS operation, your session has expired, '
                                'please use get keys to refresh your session') from error
            raise error

    def is_local(self):
        """
        Checks if stage exists
        """
        return not self._stage

    def get_decrypted_cognito_profile_id(self, profile_num=1):
        """
        Gets the cognito Account ID
        """
        return self.invoker.get_decrypted_cognito_profile_id(profile_num)

    def get_cognito_id_token(self):
        """
        Gets the cognito JWT
        """
        return self.invoker.get_cognito_id_token()

    def delete(self, event, resource='default', **kwargs):
        """
        Delete request
        """
        return self.invoker.delete(event, resource, **kwargs)

    def get(self, event, resource='default', **kwargs):
        """
        Get request
        """
        return self.invoker.get(event, resource, **kwargs)

    def post(self, event, resource='default', **kwargs):
        """
        Post request
        """
        return self.invoker.post(event, resource, **kwargs)

    def put(self, event, resource='default', **kwargs):
        """
        Put request
        """
        return self.invoker.put(event, resource, **kwargs)

    def patch(self, event, resource='default', **kwargs):
        """
        Patch request
        """
        return self.invoker.patch(event, resource, **kwargs)

    def invoke(self, event):
        """
        Direct Lambda invocation
        """
        return self.invoker.invoke(event)


def _get_boto3_error(error):
    """
    must be an instance of botocore.exceptions ClientError.
    """
    aws = {}
    message = ''

    # boto3 response
    if hasattr(error, 'response'):
        meta = error.response.get('ResponseMetadata')
        if meta:
            aws['status'] = meta['HTTPStatusCode']
        error = error.response.get('Error')
        if error:
            aws['code'] = error['Code']
            aws['message'] = error['Message']

    if hasattr(error, 'operation_name'):
        aws['op'] = error.operation_name

    if hasattr(error, 'message') or hasattr(error, 'Message'):
        message = error.message
    elif aws.get('message') or aws.get('Message'):
        message = aws['message']

    return {'message': message, 'aws': aws}



from wavelength_test.integration.cognito.cognito_client import (CognitoUserClient,
                                                         DEFAULT_ENTITLEMENTS)


def get_cognito_id_token(service_name, region, ssm_client):
    """
    Creates a new Cognito user and retrieves their JWT 'IdToken'
    :return: JWT 'IdToken'
    """
    pool_id_response = ssm_client.get_parameter(Name='/cognito/wavelength/user_pool_id')
    app_id_response = ssm_client.get_parameter(Name='/cognito/wavelength/web_client_id')
    pool_id = pool_id_response['Parameter']['Value']
    app_id = app_id_response['Parameter']['Value']

    cognito_client = CognitoUserClient(region, pool_id, app_id)
    cognito_user_data = {
        'email': f'test_{service_name}@yopmail.com',
        'first_name': f'first_name_{service_name}',
        'last_name': f'last_name_{service_name}',
        'password': 'Password1*',
        'locale': 'en_US',
        'entitlements': DEFAULT_ENTITLEMENTS
    }
    cognito_user = cognito_client.create_user(cognito_user_data)
    cognito_token = cognito_user['AuthenticationResult']['IdToken']
    return cognito_token


def get_api_key(service_name, ssm_client, gateway_client):
    """
    Get the apikeyID according to service name
    :param service_name:
    :return:
    """
    response = ssm_client.get_parameter(
        Name=f'/apigateway/{service_name}/apikeyID/default'
    )
    logical_id = response['Parameter']['Value']
    response = gateway_client.get_api_key(
        apiKey=logical_id,
        includeValue=True
    )
    return response['value']



# -*- coding: utf-8 -*-
"""
    CognitoUserClient
"""
from contextlib import suppress
import json
import boto3

DEFAULT_ENTITLEMENTS = ['*']


class CognitoUserClient:
    """
        Standard client to automate user creation/authorization in Cognito
    """

    def __init__(self, region, user_pool_id, app_client_id):
        """
        Cognito values are injected from lambda_util.py from SSM
        :param region:
        :param user_pool_id:
        :param app_client_id:
        """
        self.region = region
        self.user_pool_id = user_pool_id
        self.client_id = app_client_id
        self.client = boto3.client('cognito-idp', region_name=self.region)

    def create_user(self, user_data):
        """
            Creates a new user with NEW_PASSWORD_REQUIRED status in the specified user pool.
            If the user exists already, except is triggered and the user is logged in to get JWT
            :param user_data: dict containing user information
            :return: a dictionary containing tokens and Cognito user data
        """

        with suppress(self.client.exceptions.UsernameExistsException):
            attributes = self._get_attributes(user_data)
            attrib_set = {attr['Name']: attr['Value'] for attr in attributes}
            user = None
            try:
                user = self.client.admin_get_user(
                    UserPoolId=self.user_pool_id,
                    Username=user_data['email']
                )
            except self.client.exceptions.UserNotFoundException:
                pass
            if user:
                user_attr_set = {attr['Name']: attr['Value'] for attr in user['UserAttributes']}
                user_attr_set.pop('sub')
                if user_attr_set != attrib_set:
                    self.client.admin_update_user_attributes(
                        UserPoolId=self.user_pool_id,
                        Username=user_data['email'],
                        UserAttributes=attributes
                    )
                return self.sign_in_user(user_data['email'], user_data['password'])

            user = self.client.admin_create_user(
                UserPoolId=self.user_pool_id,
                Username=user_data['email'],
                TemporaryPassword=user_data['password'] + 'temp',
                UserAttributes=attributes
            )
            return self._authorize_new_user(user['User']['Username'], user_data['password'])

    @staticmethod
    def _get_attributes(user_data):
        return [
            {
                'Name': 'email_verified',
                'Value': 'true'
            },
            {
                'Name': 'locale',
                'Value': user_data['locale']
            },
            {
                'Name': 'given_name',
                'Value': user_data['first_name']
            },
            {
                'Name': 'family_name',
                'Value': user_data['last_name']
            },
            {
                'Name': 'email',
                'Value': user_data['email']
            },
            {
                'Name': "custom:eids",
                'Value': json.dumps({'eids': user_data['entitlements']})
            }
        ]

    def _authorize_new_user(self, user_name, password):
        """
            Automates the authorization of a new user with a password challenge/answer and gets their tokens
            :param user_data: {} containing user data
            :return: {} containing Cognito user info and tokens
        """

        auth_params = {
            'USERNAME': user_name,
            'PASSWORD': password + 'temp'
        }

        challenges = {
            'USERNAME': user_name,
            'NEW_PASSWORD': password
        }

        # do a first time login to obtain the challenge session
        response = self.client.initiate_auth(AuthFlow='USER_PASSWORD_AUTH', AuthParameters=auth_params,
                                             ClientId=self.client_id)

        session = response['Session']

        # change user password to automate the status transitioning to CONFIRMED
        auth_response = self.client.respond_to_auth_challenge(ClientId=self.client_id,
                                                              ChallengeName='NEW_PASSWORD_REQUIRED',
                                                              Session=session, ChallengeResponses=challenges)

        # no need to return user information, its in the IdToken
        return auth_response

    def sign_in_user(self, username, password):
        """
        Signs in existing user with username(email) and password and get new JWTs
        :param username:
        :param password:
        :return: {} containing JWT tokens
        """

        return self.client.admin_initiate_auth(
            UserPoolId=self.user_pool_id,
            ClientId=self.client_id,
            AuthFlow='ADMIN_NO_SRP_AUTH',
            AuthParameters={
                'USERNAME': username,
                'PASSWORD': password
            }
        )

    def sign_out_user(self, username):
        """
        Signs out users from all devices, as an administrator.
        :param username: email address of user
        :return: {}
        """

        return self.client.admin_user_global_sign_out(
            UserPoolId=self.user_pool_id,
            Username=username
        )

    def delete_user(self, email):
        """
            Deletes a user as an administrator. Works on any user.
            :param email: String
            :return: {} containing HTTPStatusCode
        """

        try:
            return self.client.admin_delete_user(
                UserPoolId=self.user_pool_id,
                Username=email
            )
        except self.client.exceptions.UserNotFoundException:
            raise Exception('An account with the given email does not exist.')

    def get_user(self, email):
        """
            Gets the specified user by user name in a user pool as an administrator. Works on any user.
            :param email: String
            :return: {} containing all user data
        """

        try:
            return self.client.admin_get_user(
                Username=email,
                UserPoolId=self.user_pool_id
            )
        except self.client.exceptions.UserNotFoundException:
            raise Exception('An account with the given email does not exist.')



"""
Event conversion utils
"""
from urllib import parse

import re

from wavelength_test.integration.util.mock_constants import (
    MOCK_TOKEN, STANDARD_AUTHORIZER_CONTEXT)

EXPECTED_HEADERS = {
    "Access-Control-Allow-Origin": "*",
    "Access-Control-Allow-Credentials": "true"
}
ALLOW_ORIGIN_OPEN_FIT_REGEX = r'.*wavelengthdev.biz'  # TODO make config
VERB_GET = 'get'
VERB_DELETE = 'delete'
VERB_POST = 'post'
VERB_PUT = 'put'
VERB_PATCH = 'patch'
VERB_INVOKE = 'invoke'

VERBS = [VERB_GET, VERB_DELETE, VERB_POST, VERB_PUT, VERB_PATCH, VERB_INVOKE]


def event_to_request_url(base_path, event):
    """
    Returns path from received event
    :param base_path:
    :param event:
    :return: path
    """

    path = base_path
    path_parameters = event.get('pathParameters')
    querystring_parameters = event.get('queryStringParameters')

    if path_parameters:
        for path_parameter in path_parameters:
            path = path.replace('{' + path_parameter + '}',
                                path_parameters.get(path_parameter))

    if querystring_parameters:
        path += '?' + parse.urlencode(querystring_parameters)

    return path


def event_to_body(event):
    """
    Gets the body from the event
    :param event:
    :return: body
    """

    body = event.get('body')
    return body


def update_request_event(event, cognito_id_token):
    """
    Updates an request event to have the request context populated with an authorization
    """
    if cognito_id_token == MOCK_TOKEN:
        # This a local call, update the event for the lambda
        if 'requestContext' not in event:
            # This mocks up the request to look like the authorizer has decomposed the id token
            event.update(STANDARD_AUTHORIZER_CONTEXT)
        elif 'authorizer' not in event['requestContext']:
            event['requestContext']['authorizer'] = STANDARD_AUTHORIZER_CONTEXT['requestContext']['authorizer']
        if 'headers' not in event:
            event['headers'] = {}
        if not event['headers'].get('Authorization'):
            event['headers']['Authorization'] = cognito_id_token

    return event


def event_to_request_headers(event, content_type, api_key, cognito_id_token):
    """
    Gets the headers from the event
    :param event:
    :param content_type:
    :param api_key:
    :param cognito_id_token: cognito jwt ('IdToken' not 'AccessToken')
    :return: headers
    """

    headers = event.get('headers', {})
    if 'Content-Type' not in headers:
        headers['Content-Type'] = content_type
    if 'x-api-key' not in headers:
        headers['x-api-key'] = api_key
    if 'Authorization' not in headers:
        headers['Authorization'] = cognito_id_token
    return headers


def check_headers(headers):
    """
    Checks for Access-Control headers
    :param headers:
    :return:
    """
    if headers:
        for expected_header_key in EXPECTED_HEADERS:
            returned_header = headers.get(expected_header_key)
            header = str(returned_header).lower()
            if header != EXPECTED_HEADERS.get(expected_header_key):
                if expected_header_key == 'Access-Control-Allow-Origin' and \
                        re.match(ALLOW_ORIGIN_OPEN_FIT_REGEX, header):
                    continue
                else:
                    raise Exception(
                        "Response does not contain expected CORS headers")
